@@ -16,7 +16,6 @@ ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 '''
 
-import OpenSSL.crypto
 import cryptography.hazmat.backends
 import cryptography.hazmat.primitives.asymmetric.rsa
 import cryptography.hazmat.primitives.hashes
@@ -25,60 +24,46 @@ import cryptography.x509.oid
 import datetime
 import os
 import requests.adapters
+import secrets
 import ssl
 import tempfile
-import urllib3.contrib.pyopenssl
 
 try:
-    from ssl import PROTOCOL_TLS as default_ssl_protocol
+    from ssl import PROTOCOL_TLS_CLIENT as default_ssl_protocol
 except ImportError:
     from ssl import PROTOCOL_SSLv23 as default_ssl_protocol
 
-def check_cert_not_after(cert):
-    cert_not_after = cert.not_valid_after
-    if cert_not_after < datetime.datetime.utcnow():
+def _check_cert_not_after(cert):
+    cert_not_after = cert.not_valid_after_utc
+    if cert_not_after < datetime.datetime.now(datetime.timezone.utc):
         raise ValueError('Client certificate expired: Not After: {cert_not_after:%Y-%m-%d %H:%M:%SZ}'.format(**locals()))
 
-def create_pyopenssl_sslcontext(pkcs12_data, pkcs12_password_bytes, ssl_protocol=default_ssl_protocol):
+def _create_sslcontext(pkcs12_data, pkcs12_password_bytes, ssl_protocol):
     private_key, cert, ca_certs = cryptography.hazmat.primitives.serialization.pkcs12.load_key_and_certificates(
         pkcs12_data,
         pkcs12_password_bytes
     )
-    check_cert_not_after(cert)
-    ssl_context = urllib3.contrib.pyopenssl.PyOpenSSLContext(ssl_protocol)
-    ssl_context._ctx.use_certificate(OpenSSL.crypto.X509.from_cryptography(cert))
-    if ca_certs:
-        for ca_cert in ca_certs:
-            check_cert_not_after(ca_cert)
-            ssl_context._ctx.add_extra_chain_cert(OpenSSL.crypto.X509.from_cryptography(ca_cert))
-    ssl_context._ctx.use_privatekey(OpenSSL.crypto.PKey.from_cryptography_key(private_key))
-    return ssl_context
-
-def create_ssl_sslcontext(pkcs12_data, pkcs12_password_bytes, ssl_protocol=default_ssl_protocol):
-    private_key, cert, ca_certs = cryptography.hazmat.primitives.serialization.pkcs12.load_key_and_certificates(
-        pkcs12_data,
-        pkcs12_password_bytes
-    )
-    check_cert_not_after(cert)
+    _check_cert_not_after(cert)
     ssl_context = ssl.SSLContext(ssl_protocol)
     with tempfile.NamedTemporaryFile(delete=False) as c:
         try:
+            tmp_pkcs12_password_bytes = secrets.token_bytes(128//8)
             pk_buf = private_key.private_bytes(
                 cryptography.hazmat.primitives.serialization.Encoding.PEM,
                 cryptography.hazmat.primitives.serialization.PrivateFormat.TraditionalOpenSSL,
-                cryptography.hazmat.primitives.serialization.BestAvailableEncryption(password=pkcs12_password_bytes)
+                cryptography.hazmat.primitives.serialization.BestAvailableEncryption(password=tmp_pkcs12_password_bytes)
             )
             c.write(pk_buf)
             buf = cert.public_bytes(cryptography.hazmat.primitives.serialization.Encoding.PEM)
             c.write(buf)
             if ca_certs:
                 for ca_cert in ca_certs:
-                    check_cert_not_after(ca_cert)
+                    _check_cert_not_after(ca_cert)
                     buf = ca_cert.public_bytes(cryptography.hazmat.primitives.serialization.Encoding.PEM)
                     c.write(buf)
             c.flush()
             c.close()
-            ssl_context.load_cert_chain(c.name, password=pkcs12_password_bytes)
+            ssl_context.load_cert_chain(c.name, password=tmp_pkcs12_password_bytes)
         finally:
             os.remove(c.name)
     return ssl_context
@@ -89,21 +74,27 @@ class Pkcs12Adapter(requests.adapters.HTTPAdapter):
         pkcs12_data = kwargs.pop('pkcs12_data', None)
         pkcs12_filename = kwargs.pop('pkcs12_filename', None)
         pkcs12_password = kwargs.pop('pkcs12_password', None)
-        ssl_protocol = kwargs.pop('ssl_protocol', default_ssl_protocol)
+        ssl_protocol_or_none = kwargs.pop('ssl_protocol', None)
         if pkcs12_data is None and pkcs12_filename is None:
             raise ValueError('Both arguments "pkcs12_data" and "pkcs12_filename" are missing')
         if pkcs12_data is not None and pkcs12_filename is not None:
             raise ValueError('Argument "pkcs12_data" conflicts with "pkcs12_filename"')
-        if pkcs12_password is None:
-            raise ValueError('Argument "pkcs12_password" is missing')
         if pkcs12_filename is not None:
             with open(pkcs12_filename, 'rb') as pkcs12_file:
                 pkcs12_data = pkcs12_file.read()
-        if isinstance(pkcs12_password, bytes):
+        if pkcs12_password is None:
+            pkcs12_password_bytes = None
+        elif isinstance(pkcs12_password, bytes):
             pkcs12_password_bytes = pkcs12_password
-        else:
+        elif isinstance(pkcs12_password, str):
             pkcs12_password_bytes = pkcs12_password.encode('utf8')
-        self.ssl_context = create_pyopenssl_sslcontext(pkcs12_data, pkcs12_password_bytes, ssl_protocol)
+        else:
+            raise TypeError('Password must be a None, string or bytes.')
+        if ssl_protocol_or_none is None:
+            ssl_protocol = default_ssl_protocol
+        else:
+            ssl_protocol = ssl_protocol_or_none
+        self.ssl_context = _create_sslcontext(pkcs12_data, pkcs12_password_bytes, ssl_protocol)
         super(Pkcs12Adapter, self).__init__(*args, **kwargs)
 
     def init_poolmanager(self, *args, **kwargs):
@@ -116,11 +107,29 @@ class Pkcs12Adapter(requests.adapters.HTTPAdapter):
             kwargs['ssl_context'] = self.ssl_context
         return super(Pkcs12Adapter, self).proxy_manager_for(*args, **kwargs)
 
+    def cert_verify(self, conn, url, verify, cert):
+        check_hostname = self.ssl_context.check_hostname
+        try:
+            if verify is False:
+                self.ssl_context.check_hostname = False
+            return super(Pkcs12Adapter, self).cert_verify(conn, url, verify, cert)
+        finally:
+            self.ssl_context.check_hostname = check_hostname
+
+    def send(self, request, stream=False, timeout=None, verify=True, cert=None, proxies=None):
+        check_hostname = self.ssl_context.check_hostname
+        try:
+            if verify is False:
+                self.ssl_context.check_hostname = False
+            return super(Pkcs12Adapter, self).send(request, stream, timeout, verify, cert, proxies)
+        finally:
+            self.ssl_context.check_hostname = check_hostname
+
 def request(*args, **kwargs):
     pkcs12_data = kwargs.pop('pkcs12_data', None)
     pkcs12_filename = kwargs.pop('pkcs12_filename', None)
     pkcs12_password = kwargs.pop('pkcs12_password', None)
-    ssl_protocol = kwargs.pop('ssl_protocol', default_ssl_protocol)
+    ssl_protocol = kwargs.pop('ssl_protocol', None)
     if pkcs12_data is None and pkcs12_filename is None and pkcs12_password is None:
         return requests.request(*args, **kwargs)
     if 'cert' in  kwargs:
@@ -159,7 +168,38 @@ def post(*args, **kwargs):
 def put(*args, **kwargs):
     return request('put', *args, **kwargs)
 
-def selftest():
+def _create_test_cert_pkcs12(key, cert, pkcs12_password_for_creation):
+    if pkcs12_password_for_creation is None:
+        algorithm = cryptography.hazmat.primitives.serialization.NoEncryption()
+    else:
+        algorithm = cryptography.hazmat.primitives.serialization.BestAvailableEncryption(pkcs12_password_for_creation)
+    pkcs12_data = cryptography.hazmat.primitives.serialization.pkcs12.serialize_key_and_certificates(
+        name=b'test',
+        key=key,
+        cert=cert,
+        cas=[cert, cert, cert],
+        encryption_algorithm=algorithm
+    )
+    return pkcs12_data
+
+def _execute_test_case(test_case_name, key, cert, pkcs12_password_for_creation, pkcs12_password, expected_status_code, expected_exception_message):
+    print(f"Testing {test_case_name}")
+    try:
+        pkcs12_data = _create_test_cert_pkcs12(key, cert, pkcs12_password_for_creation)
+        response = get(
+            'https://example.com/',
+            pkcs12_data=pkcs12_data,
+            pkcs12_password=pkcs12_password
+        )
+        if response.status_code != expected_status_code:
+            raise Exception('Unexpected response: {response!r}'.format(**locals()))
+        if expected_exception_message is not None:
+            raise Exception('Missing expected exception: {expected_exception_message!r}'.format(**locals()))
+    except ValueError as e:
+        if expected_exception_message is None or str(e) != expected_exception_message:
+            raise(e)
+
+def test():
     key = cryptography.hazmat.primitives.asymmetric.rsa.generate_private_key(public_exponent=65537, key_size=4096)
     cert = cryptography.x509.CertificateBuilder().subject_name(
         cryptography.x509.Name([
@@ -182,21 +222,9 @@ def selftest():
         cryptography.hazmat.primitives.hashes.SHA512(),
         cryptography.hazmat.backends.default_backend()
     )
-    pkcs12_data = cryptography.hazmat.primitives.serialization.pkcs12.serialize_key_and_certificates(
-        name=b'test',
-        key=key,
-        cert=cert,
-        cas=[cert, cert, cert],
-        encryption_algorithm=cryptography.hazmat.primitives.serialization.BestAvailableEncryption(b'correcthorsebatterystaple')
-    )
-    response = get(
-        'https://example.com/',
-        pkcs12_data=pkcs12_data,
-        pkcs12_password='correcthorsebatterystaple'
-    )
-    if response.status_code != 200:
-        raise Exception('Unexpected response: {response!r}'.format(**locals()))
-    print('Selftest succeeded.')
-
-if __name__ == '__main__':
-    selftest()
+    _execute_test_case('with encryption, password provided as bytes', key, cert, b'correcthorsebatterystaple', b'correcthorsebatterystaple', 200, None)
+    _execute_test_case('with encryption, password provided as string', key, cert, b'correcthorsebatterystaple', 'correcthorsebatterystaple', 200, None)
+    _execute_test_case('with empty password provided as bytes', key, cert, None, b'', 200, None)
+    _execute_test_case('with empty password provided as string', key, cert, None, '', 200, None)
+    _execute_test_case('without encryption', key, cert, None, None, 200, None)
+    print('All tests succeeded.')
